#include <avr/io.h>
#include <util/delay.h>
#include <math.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <string.h>
#include "nokia5110.h"
#include "notes.h"

#define DEFAULT_VOLUME 100
#define NUM_PLATFORMS 50
#define NUM_ENEMIES 2

ISR(TIMER1_OVF_vect)
{
    PORTD |= (1 << PD4);
	PORTD |= (1 << PD7);
}

ISR(TIMER1_COMPB_vect)
{
    PORTD &= ~(1 << PD4);
	PORTD &= ~(1 << PD7);
}

/* platforms info */
int platform_x[NUM_PLATFORMS];
int platform_y[NUM_PLATFORMS];
int platform_length[NUM_PLATFORMS];
int platform_height[NUM_PLATFORMS];
int platform_type[NUM_PLATFORMS];
int platform_lifted[NUM_PLATFORMS];

/* enemies info */
float real_x_enemy[NUM_ENEMIES];
float real_y_enemy[NUM_ENEMIES];
float y_accel_enemy[NUM_ENEMIES];
float dead_enemy[NUM_ENEMIES];
float x_speed_enemy[NUM_ENEMIES];
int go_reverse[NUM_ENEMIES];

/* player info */
float real_x;
float real_y;
int walking_left = 0;
int jumping = 0;
int grounded = 0;
double y_accel = 0.0f;
double move_speed = 0.8f;
double air_move_speed = 0.5f;

/* game states */
int dead = 0;
int should_die_soon = 0;
int lives = 3;
int coins = 0;
int finished = 0;

/* collision info */
int was_not_restricted = 1;
int between_platforms = 1;
int restrict_right = 0;
int restrict_left = 0;
int restrict_top = 0;
int descending = 0;

/* music info */
int notes_index = 0;
int note_duration = 20;
int current_block = 0;
int current_left_note = 0;
int play_coin_sound = 0;
int restarted_sound = 0;
int finished_dead_sound = 0;

/************************************************************************/
/* function that plays the Mario theme or sounds according to the game state */
/************************************************************************/
void play_music()
{
	int *notes_left;
	int *notes_right;
	
	/* the music is played using blocks of 36 notes and has 2 channels */
	switch(current_block) {
		case 0: notes_left = mario0_left; notes_right = mario0_right; break;
		case 1: notes_left = mario1_left; notes_right = mario1_right; break;
		case 2: notes_left = mario2_left; notes_right = mario2_right; break;
		case 3:	notes_left = mario1_left; notes_right = mario1_right; break;
		case 4: notes_left = mario2_left; notes_right = mario2_right; break;
		case 5: notes_left = mario3_left; notes_right = mario3_right; break;
		case 6: notes_left = mario4_left; notes_right = mario4_right; break;
		case 7: notes_left = mario3_left; notes_right = mario3_right; break;
		case 8: notes_left = mario5_left; notes_right = mario5_right; break;
		case 9: notes_left = mario3_left; notes_right = mario3_right; break;
		case 10: notes_left = mario4_left; notes_right = mario4_right; break;
		case 11: notes_left = mario3_left; notes_right = mario3_right; break;
		case 12: notes_left = mario5_left; notes_right = mario5_right; break;
		case 13: notes_left = mario6_left; notes_right = mario6_right; break;
		case 14: notes_left = mario7_left; notes_right = mario7_right; break;
		case 15: notes_left = mario6_left; notes_right = mario6_right; break;
		case 16: notes_left = mario0_left; notes_right = mario0_right; break;
		case 17: notes_left = mario1_left; notes_right = mario1_right; break;
		case 18: notes_left = mario2_left; notes_right = mario2_right; break;
		case 19: notes_left = mario1_left; notes_right = mario1_right; break;
		case 20: notes_left = mario2_left; notes_right = mario2_right; break;
		case 21: notes_left = mario8_left; notes_right = mario8_right; break;
		case 22: notes_left = mario9_left; notes_right = mario9_right; break;
		case 23: notes_left = mario8_left; notes_right = mario8_right; break;
		case 24: notes_left = mario10_left; notes_right = mario10_right; break;
		case 25: notes_left = mario8_left; notes_right = mario8_right; break;
		case 26: notes_left = mario9_left; notes_right = mario9_right; break;
		case 27: notes_left = mario8_left; notes_right = mario8_right; break;
		case 28: notes_left = mario10_left; notes_right = mario10_right; break;
		case 29: notes_left = mario6_left; notes_right = mario6_right; break;
		case 30: notes_left = mario7_left; notes_right = mario7_right; break;
		case 31: notes_left = mario6_left; notes_right = mario6_right; break;
		case 32: notes_left = mario0_left; notes_right = mario0_right; break;
		case 33: notes_left = mario8_left; notes_right = mario8_right; break;
		case 34: notes_left = mario9_left; notes_right = mario9_right; break;
		case 35: notes_left = mario8_left; notes_right = mario8_right; break;
		case 36: notes_left = mario10_left; notes_right = mario10_right; break;
		default:
			current_block = 0; break;
	}

	int note_left, note_right;
	
	// in case Mario dies but game is not over, override the music with the death music
	if(dead == 1 && lives > 1) 
		notes_left = notes_right = mario_death;

	// in case Mario dies and game is over, override the music with the game over music
	if(dead == 1 && lives == 1) 
		notes_left = notes_right = game_over;
		
	// when Mario collects a star, play the star sound only on left channel
	if(play_coin_sound == 1) 
		notes_left = star_sound;
		
	// if level is finished, print a message and override the sounds with the castle music
	if(finished == 1) {
		nokia_lcd_set_cursor(45, 20);
		nokia_lcd_write_string("GOOD", 1);
		nokia_lcd_set_cursor(47, 28);
		nokia_lcd_write_string("JOB", 1);
		notes_left = notes_right = castle;
	}			
	
	// read the notes using program memory
	note_left = pgm_read_word_near(&(notes_left[notes_index]));
	note_right = pgm_read_word_near(&(notes_right[notes_index]));
	
	if(play_coin_sound == 1)
		note_left = pgm_read_word_near(&(notes_left[current_left_note]));

	if(note_left == MUSIC_END && play_coin_sound) {
		current_left_note = 0;
		play_coin_sound = 0;
	}
	
	// if a block finishes, go to the next or print a message if dead
	if(note_right == MUSIC_END) {
		if(dead == 1) {
			// mute both channels
			OCR1B = 0;
			OCR2A = 0;
			OCR2B = 0;
			TCCR2A = 0;
			TCCR2B = 0;
			finished_dead_sound = 1;
			nokia_lcd_set_cursor(2, 10);
			if(lives > 1) 
				nokia_lcd_write_string("PRESS ANY KEY", 1);
			else
				nokia_lcd_write_string("GAME OVER", 1);
			return;
		}			
		
		current_block++;
		notes_index = 0;
		play_music(current_block);
		return;						
	}	
			
	
	// on note read on left channel, play it
	if(p == note_left)
		OCR1B = 0;
	else {
		OCR1B = DEFAULT_VOLUME;
		ICR1H = (note_left >> 8);
		ICR1L = note_left;
	}
	
	// on note read on right channel, play it
	if(p == note_right) {
		OCR2A = 0;
		OCR2B = 0;
		TCCR2A = 0;
		TCCR2B = 0;	
	} else {
		TCCR2A = (1<<WGM21) | (1<<COM2A0);
		TCCR2B = (7<<CS20);
		OCR2A = 16000000 / 2048 / (1000000 / note_right);
	}

	// count the duration of a note
	if(note_duration == 0) { 
			if(play_coin_sound == 1)
				current_left_note++;
			notes_index++;
			note_duration = 2;	
	}			
	else 
		note_duration--;
}


/************************************************************************/
/* player move left function; slower midair                             */
/************************************************************************/
void move_left() {
	if(!restrict_left)
		if(!grounded)
			real_x -= air_move_speed;
		else
			real_x -= move_speed;
		
	walking_left = 1;
}

/************************************************************************/
/* player move right function; slower midair                            */
/************************************************************************/
void move_right() {
	if(!restrict_right) 
		if(!grounded)
			real_x += air_move_speed;
		else
			real_x += move_speed;
		
	walking_left = 0;
}

/************************************************************************/
/* player jump function; movement on y-axis is accelerated              */
/************************************************************************/
void jump() {
	if(grounded) {
		jumping = 1;
		y_accel = 1.7f;
		grounded = 0;
	}
}

/************************************************************************/
/* checks the current state of the jump: jumping or falling              */
/************************************************************************/
void check_jumping() {
	if(jumping) {
		real_y -= y_accel;
		y_accel -= 0.08f;
		debug("sare");
	}		
	if(y_accel < 0) {
		jumping = 0;
		y_accel = 0;
	}		
}

/************************************************************************/
/* level initialization - platforms and enemies                         */
/************************************************************************/
void init_level() {
	int i;
	for(i = 0; i < NUM_PLATFORMS; i++) {
		platform_x[i] = 0;
		platform_y[i] = 0;
		platform_length[i] = 0;
		platform_type[i] = -1;
		platform_lifted[i] = 0;
	}	
	
	real_x_enemy[0] = 74;
	real_y_enemy[0] = 15;
	y_accel_enemy[0] = 0;
	dead_enemy[0] = 0;
	x_speed_enemy[0] = 0.3f;
}

/************************************************************************/
/* player initialization                                                */
/************************************************************************/
void init_player() {
	real_x = 30.0f;
	real_y = 10.0f;
	finished = 0;
}

/************************************************************************/
/* renders the enemies specified in the arrays                          */
/************************************************************************/
void render_enemies() {
	int i;
	for(i = 0; i < NUM_ENEMIES - 1; i++) {
		if(dead_enemy[i])
			return;
		
		if(dead)
			x_speed_enemy[i] = 0;
		
		int scroll_x = (30 - (int) real_x);
		nokia_lcd_set_cursor((int)real_x_enemy[i] + scroll_x, (int)real_y_enemy[i]);
		nokia_lcd_write_special(102, 1, 7, 5);
	
		real_x_enemy[i] += x_speed_enemy[i];
		
		if(real_x_enemy[i] > 154)
			go_reverse[i] = 1;
		
		if(real_x_enemy[i] < 1)
			go_reverse[i] = 0;
		
		if(go_reverse[i] && (int)real_y_enemy[i] == 32)
			real_x_enemy[i] -= 2 * x_speed_enemy[i];
	
		if(real_x_enemy[i] > 114 && real_y_enemy[i] < 32) {
			y_accel_enemy[i] += 0.1f;
			real_y_enemy[i] += y_accel_enemy[i];
		}
	
		if(real_y_enemy[i] >= 32) {
			y_accel_enemy[i] = 0;
		}
	}	
	
}



void debug(const char *mesaj) {
	//nokia_lcd_set_cursor(0,0);
	//nokia_lcd_write_string(mesaj, 1);
}


/************************************************************************/
/* function responsible for collision checking                          */
/************************************************************************/
void check_collision() {
	int i;
	
	// states 
	grounded = 0;
	restrict_left = 0;
	restrict_right = 0;
	
	// if the player goes lower than the lowest platform, it means he's dead
	if(real_y + 5 > platform_y[0]) {
		should_die_soon = 1;
		real_y += y_accel;
		y_accel += 0.03f;
		if(real_y + 7 > 65)	{
			notes_index = 0;
			dead = 1;
		}
		return;
	}					
	
	// finish flag collision and animation
	if(real_x > 380 && platform_lifted[35] < 24) {
		platform_lifted[35]++;
		finished = 1;
		if(!restarted_sound) {
			notes_index = 0;
			restarted_sound = 1;
			coins += 50;
		}			
	}	
	
	// collision check for enemies
	// hitting them from top results in killing them and giving coins
	// hitting them some other way results in dying
	for(i = 0; i < NUM_ENEMIES; i++) {
		if((int)real_y + 7 == (int)real_y_enemy[i] && 
			(int)real_x + 5 >= (int)real_x_enemy[i] && 
			(int)real_x < (int)real_x_enemy[i] + 5 && 
			!dead_enemy[i]) {
				dead_enemy[0] = 1;
				coins += 30;
		}
		
		if((int)real_y + 7 > (int)real_y_enemy[i] && 
			(int)real_x + 5 == (int)real_x_enemy[i] && 
			(int)real_y < (int)real_y_enemy[i] + 7 && 
			!dead_enemy[i])
				dead = 1;
			
		if((int)real_y + 7 > (int)real_y_enemy[i] && 
			(int)real_x == (int)real_x_enemy[i] + 5  && 
			(int)real_y < (int)real_y_enemy[i] + 7 && 
			!dead_enemy[i])
				dead = 1;
	}	
	
	// block player if he goes too far to the left
	if(real_x <= 1)
		restrict_left = 1;
		
	// check platform collision
	for(i = 0; i < NUM_PLATFORMS; i++) {
		// collision from left => restricts right movement
		if((int)real_y + 7 > platform_y[i] && 
			(int)real_x + 6 == platform_x[i] && 
			(int)real_y < platform_y[i] + platform_height[i])
				restrict_right = 1;
	
		// collision from right => restrict left movement
		if((int)real_y + 7 > platform_y[i] && 
			(int)real_x == platform_x[i] + platform_length[i]  && 
			(int)real_y < platform_y[i] + platform_height[i])
				restrict_left = 1;
			
		// collision from top => player is grounded
		if((int)real_x + 4 >= platform_x[i] && 
			(int)real_x < platform_x[i] + platform_length[i] && 
			was_not_restricted == 1) {
				if(platform_height[i] + platform_y[i] - real_y < 1) {
					restrict_top = 1;
					was_not_restricted = 0;
					jumping = 0;
					y_accel = 0;
					real_y = platform_height[i] + platform_y[i];
					if(platform_type[i] == 1) {
						platform_type[i] = 0;
						coins += 10;
						play_coin_sound = 1;
					}
					platform_lifted[i] += 1;
			}			
		}
		
		// limit jumping so the player doesn't go out of the screen
		if((int)real_y == 0) {
			restrict_top = 1;
			was_not_restricted = 0;
			jumping = 0;
			y_accel = 0;
			real_y = 1;
		}	
		
		// hitting the bricks with the head from below results in lifting them
		if(platform_length[i] == 0)
			break;
		else 
			if(((int)real_x + 5 > platform_x[i] 
				&& (int)real_x < platform_x[i] + platform_length[i])  
				&& !jumping 
				&& !should_die_soon) {
					if(real_y + 7 - platform_y[i] < 2 
						&& real_y + 7 - platform_y[i] >= 0) {
							debug("ground");
							grounded = 1;
							y_accel = 0;
							restrict_top = 0;
							was_not_restricted = 1;
							real_y = platform_y[i] - 7;
							int j;
						
							// update the lifted values
							for(j = 0; j < NUM_PLATFORMS; j++)
								if(platform_lifted[j] != 0 && platform_type[j] != 2)
									platform_lifted[j] = 0;
							break;
					}						
				} 	
	}
	
	// in case of unwanted break above, recheck left and right restrictions
	for(i = 0; i < NUM_PLATFORMS; i++) {
		if((int)real_y + 7 > platform_y[i] && 
			(int)real_x + 6 == platform_x[i] && 
			(int)real_y < platform_y[i] + platform_height[i])
				restrict_right = 1;
			
		if((int)real_y + 7 > platform_y[i] && 
			(int)real_x == platform_x[i] + platform_length[i]  && 
			(int)real_y < platform_y[i] + platform_height[i])
				restrict_left = 1;
	}
	
	// if falling, accelerate the player downward
	if(!grounded && !jumping) {
		real_y += y_accel;
		y_accel += 0.03f;
		if(real_y + 7 > 65)	{
			notes_index = 0;
			dead = 1;
		}			
		else 
			debug("cade");	
	}
}

/************************************************************************/
/* function that renders the player object - flipped or not             */
/************************************************************************/
void render_player() {
	nokia_lcd_set_cursor(30, (int)real_y);
	nokia_lcd_write_special(97 + walking_left, 1, 7, 5);
	if(!dead) {
		check_collision();
		check_jumping();
	}
}


/************************************************************************/
/* platform drawing function depending on the type						*/
/* type 0: normal platform												*/
/* type 1: platform with a coin											*/
/* type 2: finish flag													*/
/* type 3: pipe                                                         */
/************************************************************************/
void draw_platform(int x0, int y0, int x1, int y1, int index, int type) {
	platform_length[index] = x1 - x0;
	platform_x[index] = x0;
	platform_y[index] = y0;
	platform_height[index] = y1 - y0;
	if(platform_type[index] == -1)
		platform_type[index] = type;
		
	float scroll_x = 30 - real_x;
	int i, j;

	switch(platform_type[index]) {
		case -1:
			return;
		case 0:
			for (i = fmax(0, x0 + scroll_x) + 1; i < fmin(fmax(0, x1 + scroll_x), 84); i++) {
				nokia_lcd_set_cursor(i , y0 - platform_lifted[index]);
				nokia_lcd_write_special(96, 1, y1 - y0, 1);
			}				
			break;
		case 1:
			if(x1 + scroll_x < 0)
				break;
			nokia_lcd_set_cursor(x0 + scroll_x + 2, y0 - platform_lifted[index]);
			nokia_lcd_write_special(99, 1, 7, 5);
			break;
			
		case 2:
			if(x0 + scroll_x + 1 < 0 || x0 + scroll_x + 6 >= 84)
				break;
			nokia_lcd_set_cursor(x0 + scroll_x + 1, y0 - platform_lifted[index]);
			nokia_lcd_write_special(3, 1, 7, 5);
			break;
			
		case 3:
			for (i = fmax(0, x0 + scroll_x + 2); i < fmin(fmax(0, x1 + scroll_x - 1), 84); i++) {
				nokia_lcd_set_cursor(i , y0);
				nokia_lcd_write_special(96, 1, y1 - y0, 1);
			}				
			nokia_lcd_set_cursor((int)fmax(0, x0 + scroll_x + 1), y0);
			nokia_lcd_write_special(96, 1, 1, 1);
			nokia_lcd_set_cursor((int)fmin(fmax(0, x1 + scroll_x), 84), y0);
			nokia_lcd_write_special(96, 1, 1, 1);
			break;
	}				
}

/************************************************************************/
/* function for hardware IO initialization                              */
/************************************************************************/
void IO_init() {
	// initializes the Nokia display
	nokia_lcd_init();
	nokia_lcd_clear();
	nokia_lcd_backlight(1);
	
	// initializes the button ports
	DDRD &= ~(1 << PD0);
	DDRD &= ~(1 << PD1);
	DDRB &= ~(1 << PB0);
	PORTB |= (1 << PB0);
	PORTD |= (1 << PD0);
	PORTD |= (1 << PD1);
	
	// initialize the sound on timer1
	DDRD |= _BV(PD4); 
    TCCR1A |= _BV(COM1B1);

	TCCR1B |= _BV(WGM13)
		   |_BV(CS11); 	
		   
	// initialize the sound on timer2
	DDRD |= _BV(PD7);
	TCCR2A = (1<<WGM21) | (1<<COM2A0);
	TCCR2B = (7<<CS20);
		  
}

/************************************************************************/
/* resets the level                                                     */
/************************************************************************/
void reset_level() {
	init_player();
	init_level();
	dead = 0;
	y_accel = 0.0f;
	move_speed = 0.8f;
	air_move_speed = 0.5f;
	current_block = 0;
	notes_index = 0;
	finished_dead_sound = 0;
	should_die_soon = 0;
	lives--;
	go_reverse[0] = 0;
}

/************************************************************************/
/* functions that detect the pressed keys                               */
/************************************************************************/
int left_pressed() {
	return ((PIND & (1 << PD0)) == 0)?1:0;
}

int right_pressed() {
	return ((PIND & (1 << PD1)) == 0)?1:0;
}

int up_pressed() {
	return ((PINB & (1 << PB0)) == 0)?1:0;
}


int main(void) {
	IO_init();
	init_player();
	init_level();
	
	sei();
	
	while(1) {
		if (left_pressed() && !finished && !dead)
			move_left();
		
		if (right_pressed() && !finished && !dead)
			move_right();

		if (up_pressed() && !finished && !dead) 
			jump(); 
			
		if(dead && 
			(left_pressed() || right_pressed() || up_pressed()) 
				&& finished_dead_sound == 1 && lives >= 2)
			reset_level();
			
		play_music();
		
		// here the platforms are drawn and initialized
		draw_platform(0, 40, 200, 47, 0, 0);
		draw_platform(55, 22, 63, 29, 1, 0);
		
		draw_platform(74, 22, 82, 29, 2, 0);
		draw_platform(83, 22, 90, 29, 3, 1);
		draw_platform(91, 22, 98, 29, 4, 0);
		draw_platform(99, 22, 106, 29, 5, 1);
		draw_platform(107, 22, 114, 29, 6, 0);
		
		draw_platform(91, 5, 98, 12, 7, 1);
		
		draw_platform(220, 40, 281, 47, 8, 0);
		
		draw_platform(257, 35, 263, 40, 9, 0);
		draw_platform(263, 30, 269, 35, 10, 0);
		draw_platform(263, 35, 269, 40, 11, 0);
		draw_platform(269, 25, 275, 30, 12, 0);
		draw_platform(269, 30, 275, 35, 13, 0);
		draw_platform(269, 35, 275, 40, 14, 0);
		draw_platform(275, 20, 281, 25, 15, 0);
		draw_platform(275, 25, 281, 30, 16, 0);
		draw_platform(275, 30, 281, 35, 17, 0);
		draw_platform(275, 35, 281, 40, 18, 0);
		
		draw_platform(295, 40, 450, 47, 19, 0);
		draw_platform(295, 20, 301, 25, 20, 0);
		draw_platform(295, 25, 301, 30, 21, 0);
		draw_platform(295, 30, 301, 35, 22, 0);
		draw_platform(295, 35, 301, 40, 23, 0);
		draw_platform(301, 25, 307, 30, 24, 0);
		draw_platform(301, 30, 307, 35, 25, 0);
		draw_platform(301, 35, 307, 40, 26, 0);
		draw_platform(307, 30, 313, 35, 27, 0);
		draw_platform(307, 35, 313, 40, 28, 0);
		draw_platform(313, 35, 319, 40, 29, 0);
		
		draw_platform(385, 35, 392, 40, 30, 0);
		draw_platform(388, 28, 389, 35, 31, 0);
		draw_platform(388, 21, 389, 28, 32, 0);
		draw_platform(388, 14, 389, 21, 33, 0);
		draw_platform(388, 7, 389, 14, 34, 0);
		draw_platform(389, 28, 394, 35, 35, 2);
		
		draw_platform(159, 33, 167, 40, 36, 3);
		
		// print number of hearts according to current lives
		if(lives >= 1) {
			nokia_lcd_set_cursor(0, 2);
			nokia_lcd_write_special(100, 1, 7, 5);
		}
		
		if(lives >= 2) {
			nokia_lcd_set_cursor(6, 2);
			nokia_lcd_write_special(100, 1, 7, 5);
		}
		
		if(lives >= 3) {
			nokia_lcd_set_cursor(12, 2);
			nokia_lcd_write_special(100, 1, 7, 5);
		}
		
		// print number of coins and the $ symbol
		nokia_lcd_set_cursor(60, 0);
		nokia_lcd_write_special(101, 1, 7, 5);
		nokia_lcd_set_cursor(67, 0);
		char aux[4];
		sprintf(aux, "%d", coins);
		nokia_lcd_write_string(aux, 1);
		
		render_player();
		render_enemies();

		nokia_lcd_render();
		_delay_ms(10);
		
		nokia_lcd_clear();
	}
	
}
